/*
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */
package org.apache.qpid.test.testcases;

import org.apache.qpid.test.framework.AMQPPublisher;
import org.apache.qpid.test.framework.Circuit;
import org.apache.qpid.test.framework.FrameworkBaseCase;
import org.apache.qpid.test.framework.MessagingTestConfigProperties;
import static org.apache.qpid.test.framework.MessagingTestConfigProperties.*;
import org.apache.qpid.test.framework.sequencers.CircuitFactory;

import org.apache.qpid.junit.extensions.util.ParsedProperties;
import org.apache.qpid.junit.extensions.util.TestContextProperties;

/**
 * MandatoryMessageTest tests for the desired behaviour of mandatory messages. Mandatory messages are a non-JMS
 * feature. A message may be marked with a mandatory delivery flag, which means that a valid route for the message
 * must exist, when it is sent, or when its transaction is committed in the case of transactional messaging. If this
 * is not the case, the broker should return the message with a NO_CONSUMERS code.
 *
 * <p><table id="crc"><caption>CRC Card</caption>
 * <tr><th> Responsibilities <th> Collaborations
 * <tr><td> Check that an mandatory message is sent succesfully not using transactions when a consumer is connected.
 * <tr><td> Check that an mandatory message is committed succesfully in a transaction when a consumer is connected.
 * <tr><td> Check that a mandatory message is sent succesfully, not using transactions, when a consumer is disconnected
 *          but the route exists.
 * <tr><td> Check that a mandatory message is sent succesfully, in a transaction, when a consumer is disconnected but
 *          the route exists.
 * <tr><td> Check that an mandatory message results in no route code, not using transactions, when no consumer is
 *          connected.
 * <tr><td> Check that an mandatory message results in no route code, upon transaction commit, when a consumer is
 *          connected.
 * <tr><td> Check that an mandatory message is sent succesfully not using transactions when a consumer is connected.
 * <tr><td> Check that an mandatory message is committed succesfully in a transaction when a consumer is connected.
 * <tr><td> Check that a mandatory message is sent succesfully, not using transactions, when a consumer is disconnected
 *          but the route exists.
 * <tr><td> Check that a mandatory message is sent succesfully, in a transaction, when a consumer is disconnected but
 *          the route exists.
 * <tr><td> Check that an mandatory message results in no route code, not using transactions, when no consumer is
 *          connected.
 * <tr><td> Check that an mandatory message results in no route code, upon transaction commit, when a consumer is
 *          connected.
 * </table>
 *
 * @todo All of these test cases will be generated by a test generator that thoroughly tests all combinations of test
 *       circuits.
 */
public class MandatoryMessageTest extends FrameworkBaseCase
{
    /** Used to read the tests configurable properties through. */
    ParsedProperties testProps;

    /**
     * Creates a new test case with the specified name.
     *
     * @param name The test case name.
     */
    public MandatoryMessageTest(String name)
    {
        super(name);
    }

    /** Check that an mandatory message is sent succesfully not using transactions when a consumer is connected. */
    public void test_QPID_508_MandatoryOkNoTxP2P() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Run the default test sequence over the test circuit checking for no errors.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /** Check that an mandatory message is committed succesfully in a transaction when a consumer is connected. */
    public void test_QPID_508_MandatoryOkTxP2P() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Run the default test sequence over the test circuit checking for no errors.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /**
     * Check that a mandatory message is sent succesfully, not using transactions, when a consumer is disconnected but
     * the route exists.
     */
    public void test_QPID_517_MandatoryOkConsumerDisconnectedNoTxP2P() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Disconnect the consumer.
        testProps.setProperty(RECEIVER_CONSUMER_ACTIVE_PROPNAME, false);

        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        // Send one message with no errors.
        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /**
     * Check that a mandatory message is sent succesfully, in a transaction, when a consumer is disconnected but
     * the route exists.
     */
    public void test_QPID_517_MandatoryOkConsumerDisconnectedTxP2P() throws Exception
    {
        // Ensure transactional sessions are on.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Disconnect the consumer.
        testProps.setProperty(RECEIVER_CONSUMER_ACTIVE_PROPNAME, false);

        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        // Send one message with no errors.
        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /** Check that an mandatory message results in no route code, not using transactions, when no consumer is connected. */
    public void test_QPID_508_MandatoryFailsNoRouteNoTxP2P() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Set up the messaging topology so that only the publishers producer is bound (do not set up the receivers to
        // collect its messages).
        testProps.setProperty(RECEIVER_CONSUMER_BIND_PROPNAME, false);

        // Send one message and get a linked no route exception.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noRouteAssertion(testProps))));
    }

    /** Check that an mandatory message results in no route code, upon transaction commit, when a consumer is connected. */
    public void test_QPID_508_MandatoryFailsNoRouteTxP2P() throws Exception
    {
        // Ensure transactional sessions are on.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, false);

        // Set up the messaging topology so that only the publishers producer is bound (do not set up the receivers to
        // collect its messages).
        testProps.setProperty(RECEIVER_CONSUMER_BIND_PROPNAME, false);

        // Send one message and get a linked no route exception.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noRouteAssertion(testProps))));
    }

    /** Check that an mandatory message is sent succesfully not using transactions when a consumer is connected. */
    public void test_QPID_508_MandatoryOkNoTxPubSub() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Run the default test sequence over the test circuit checking for no errors.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /** Check that an mandatory message is committed succesfully in a transaction when a consumer is connected. */
    public void test_QPID_508_MandatoryOkTxPubSub() throws Exception
    {
        // Ensure transactional sessions are on.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Run the default test sequence over the test circuit checking for no errors.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /**
     * Check that a mandatory message is sent succesfully, not using transactions, when a consumer is disconnected but
     * the route exists.
     */
    public void test_QPID_517_MandatoryOkConsumerDisconnectedNoTxPubSub()  throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Use durable subscriptions, so that the route remains open with no subscribers.
        testProps.setProperty(DURABLE_SUBSCRIPTION_PROPNAME, true);

        // Disconnect the consumer.
        testProps.setProperty(RECEIVER_CONSUMER_ACTIVE_PROPNAME, false);

        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        // Send one message with no errors.
        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /**
     * Check that a mandatory message is sent succesfully, in a transaction, when a consumer is disconnected but
     * the route exists.
     */
    public void test_QPID_517_MandatoryOkConsumerDisconnectedTxPubSub() throws Exception
    {
        // Ensure transactional sessions are on.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Use durable subscriptions, so that the route remains open with no subscribers.
        testProps.setProperty(DURABLE_SUBSCRIPTION_PROPNAME, true);

        // Disconnect the consumer.
        testProps.setProperty(RECEIVER_CONSUMER_ACTIVE_PROPNAME, false);

        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        // Send one message with no errors.
        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noExceptionsAssertion(testProps))));
    }

    /** Check that an mandatory message results in no route code, not using transactions, when no consumer is connected. */
    public void test_QPID_508_MandatoryFailsNoRouteNoTxPubSub() throws Exception
    {
        // Ensure transactional sessions are off.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, false);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Set up the messaging topology so that only the publishers producer is bound (do not set up the receivers to
        // collect its messages).
        testProps.setProperty(RECEIVER_CONSUMER_BIND_PROPNAME, false);

        // Send one message and get a linked no route exception.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noRouteAssertion(testProps))));
    }

    /** Check that an mandatory message results in no route code, upon transaction commit, when a consumer is connected. */
    public void test_QPID_508_MandatoryFailsNoRouteTxPubSub() throws Exception
    {
        // Ensure transactional sessions are on.
        testProps.setProperty(TRANSACTED_PUBLISHER_PROPNAME, true);
        testProps.setProperty(PUBSUB_PROPNAME, true);

        // Set up the messaging topology so that only the publishers producer is bound (do not set up the receivers to
        // collect its messages).
        testProps.setProperty(RECEIVER_CONSUMER_BIND_PROPNAME, false);

        // Send one message and get a linked no route exception.
        CircuitFactory circuitFactory = getCircuitFactory();
        Circuit testCircuit = circuitFactory.createCircuit(getConnection(), testProps);

        assertNoFailures(testCircuit.test(1,
                assertionList(((AMQPPublisher) testCircuit.getPublisher()).noRouteAssertion(testProps))));
    }

    protected void setUp() throws Exception
    {
        super.setUp();

        testProps = TestContextProperties.getInstance(MessagingTestConfigProperties.defaults);

        /** All these tests should have the mandatory flag on. */
        testProps.setProperty(IMMEDIATE_PROPNAME, false);
        testProps.setProperty(MANDATORY_PROPNAME, true);

        /** Bind the receivers consumer by default. */
        testProps.setProperty(RECEIVER_CONSUMER_BIND_PROPNAME, true);
        testProps.setProperty(RECEIVER_CONSUMER_ACTIVE_PROPNAME, true);
    }
}
